"""Test config options."""
import logging
from unittest.mock import patch

import pytest

from theano import configdefaults, configparser
from theano.configdefaults import default_blas_ldflags
from theano.configparser import THEANO_FLAGS_DICT, AddConfigVar, ConfigParam


def test_invalid_default():
    # Ensure an invalid default value found in the Theano code only causes
    # a crash if it is not overridden by the user.

    def validate(val):
        if val == "invalid":
            raise ValueError("Test-triggered")

    with pytest.raises(ValueError, match="Test-triggered"):
        # This should raise a ValueError because the default value is
        # invalid.
        AddConfigVar(
            "T_config__test_invalid_default_a",
            doc="unittest",
            configparam=ConfigParam("invalid", validate=validate),
            in_c_key=False,
        )

    THEANO_FLAGS_DICT["T_config__test_invalid_default_b"] = "ok"
    # This should succeed since we defined a proper value, even
    # though the default was invalid.
    AddConfigVar(
        "T_config__test_invalid_default_b",
        doc="unittest",
        configparam=ConfigParam("invalid", validate=validate),
        in_c_key=False,
    )

    # TODO We should remove these dummy options on test exit.
    # Check that the flag has been removed
    assert "T_config__test_invalid_default_b" not in THEANO_FLAGS_DICT


@patch("theano.configdefaults.try_blas_flag", return_value=None)
@patch("theano.configdefaults.sys")
def test_default_blas_ldflags(sys_mock, try_blas_flag_mock, caplog):

    sys_mock.version = "3.8.0 | packaged by conda-forge | (default, Nov 22 2019, 19:11:38) \n[GCC 7.3.0]"

    with patch.dict("sys.modules", {"mkl": None}):
        with caplog.at_level(logging.WARNING):
            default_blas_ldflags()

    assert "install mkl with" in caplog.text


def test_config_param_apply_and_validation():
    cp = ConfigParam(
        "TheDeFauLt",
        apply=lambda v: v.lower(),
        validate=lambda v: v in "thedefault,thesetting",
        mutable=True,
    )
    assert cp.default == "TheDeFauLt"
    assert not hasattr(cp, "val")

    # can't assign invalid value
    with pytest.raises(ValueError, match="Invalid value"):
        cp.__set__("cls", "invalid")

    assert not hasattr(cp, "val")

    # effectivity of apply function
    cp.__set__("cls", "THESETTING")
    assert cp.val == "thesetting"

    # respect the mutability
    cp._mutable = False
    with pytest.raises(Exception, match="Can't change"):
        cp.__set__("cls", "THEDEFAULT")


def test_config_hash():
    # TODO: use custom config instance for the test
    root = configparser.config
    configparser.AddConfigVar(
        "test_config_hash",
        "A config var from a test case.",
        configparser.StrParam("test_default"),
        root=root,
    )

    h0 = configparser.get_config_hash()

    with configparser.change_flags(test_config_hash="new_value"):
        assert root.test_config_hash == "new_value"
        h1 = configparser.get_config_hash()

    h2 = configparser.get_config_hash()
    assert h1 != h0
    assert h2 == h0


class TestConfigTypes:
    def test_bool(self):
        valids = {
            True: ["1", 1, True, "true", "True"],
            False: ["0", 0, False, "false", "False"],
        }
        param = configparser.BoolParam(None)
        assert isinstance(param, configparser.ConfigParam)
        assert param.default is None
        for outcome, inputs in valids.items():
            for input in inputs:
                applied = param.apply(input)
                assert applied == outcome
                assert param.validate(applied) is not False
        with pytest.raises(ValueError, match="Invalid value"):
            param.apply("notabool")

    def test_enumstr(self):
        cp = configparser.EnumStr("blue", ["red", "green", "yellow"])
        assert len(cp.all) == 4
        with pytest.raises(ValueError, match=r"Invalid value \('foo'\)"):
            cp.apply("foo")
        with pytest.raises(ValueError, match="Non-str value"):
            configparser.EnumStr(default="red", options=["red", 12, "yellow"])

    def test_deviceparam(self):
        cp = configparser.DeviceParam("cpu", mutable=False)
        assert cp.default == "cpu"
        assert cp._apply("cuda123") == "cuda123"
        with pytest.raises(ValueError, match="old GPU back-end"):
            cp._apply("gpu123")
        with pytest.raises(ValueError, match="Invalid value"):
            cp._apply("notadevice")
        assert str(cp) == "None (cpu, opencl*, cuda*) "


def test_config_context():
    # TODO: use custom config instance for the test
    root = configparser.config
    configparser.AddConfigVar(
        "test_config_context",
        "A config var from a test case.",
        configparser.StrParam("test_default"),
        root=root,
    )
    assert hasattr(root, "test_config_context")
    assert root.test_config_context == "test_default"

    with configparser.change_flags(test_config_context="new_value"):
        assert root.test_config_context == "new_value"
    assert root.test_config_context == "test_default"


def test_no_more_dotting():
    with pytest.raises(ValueError, match="Dot-based"):
        AddConfigVar(
            "T_config.something",
            doc="unittest",
            configparam=ConfigParam("invalid"),
            in_c_key=False,
        )


def test_mode_apply():

    assert configdefaults.filter_mode("DebugMode") == "DebugMode"

    with pytest.raises(ValueError, match="Expected one of"):
        configdefaults.filter_mode("not_a_mode")

    # test with theano.Mode instance
    import theano.compile.mode

    assert (
        configdefaults.filter_mode(theano.compile.mode.FAST_COMPILE)
        == theano.compile.mode.FAST_COMPILE
    )



"""

To update the `Scan` Cython code you must
- update the version value in this file and `scan_perform.py`, and
- run `cython scan_perform.pyx; mv scan_perform.c c_code`

"""


import errno
import logging
import os
import sys
import warnings
from importlib import reload

import numpy as np

import theano
from theano import config
from theano.gof import cmodule
from theano.gof.compilelock import get_lock, release_lock


_logger = logging.getLogger("theano.scan.scan_perform")


version = 0.297  # must match constant returned in function get_version()

need_reload = False


def try_import():
    global scan_perform
    sys.path[0:0] = [config.compiledir]
    import scan_perform

    del sys.path[0]


def try_reload():
    sys.path[0:0] = [config.compiledir]
    reload(scan_perform)
    del sys.path[0]


try:
    try_import()
    need_reload = True
    if version != getattr(scan_perform, "_version", None):
        raise ImportError()
except ImportError:
    get_lock()
    try:
        # Maybe someone else already finished compiling it while we were
        # waiting for the lock?
        try:
            if need_reload:
                # The module was successfully imported earlier: we need to
                # reload it to check if the version was updated.
                try_reload()
            else:
                try_import()
                need_reload = True
            if version != getattr(scan_perform, "_version", None):
                raise ImportError()
        except ImportError:
            if not theano.config.cxx:
                raise ImportError("no c compiler, can't compile cython code")
            _logger.info("Compiling C code for scan")
            dirname = "scan_perform"
            cfile = os.path.join(theano.__path__[0], "scan", "c_code", "scan_perform.c")
            if not os.path.exists(cfile):
                # This can happen in not normal case. We just
                # disable the cython code. If we are here the user
                # didn't disable the compiler, so print a warning.
                warnings.warn(
                    "The file scan_perform.c is not available. This do"
                    "not happen normally. You are probably in a strange"
                    "setup. This mean Theano can not use the cython code for "
                    "scan. If you"
                    "want to remove this warning, use the Theano flag"
                    "'cxx=' (set to an empty string) to disable all c"
                    "code generation."
                )
                raise ImportError("The file lazylinker_c.c is not available.")

            with open(cfile) as f:
                code = f.read()
            loc = os.path.join(config.compiledir, dirname)
            if not os.path.exists(loc):
                try:
                    os.mkdir(loc)
                except OSError as e:
                    assert e.errno == errno.EEXIST
                    assert os.path.exists(loc)

            preargs = ["-fwrapv", "-O2", "-fno-strict-aliasing"]
            preargs += cmodule.GCC_compiler.compile_args()
            # Cython 19.1 always use the old NumPy interface.  So we
            # need to manually modify the .c file to get it compiled
            # by Theano. As by default, we tell NumPy to don't import
            # the old interface.
            if False:
                # During scan cython development, it is helpful to keep the old interface, to don't manually edit the c file each time.
                preargs.remove("-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION")
            else:
                numpy_ver = [int(n) for n in np.__version__.split(".")[:2]]
                # Add add some macro to lower the number of edit
                # needed to the c file.
                if bool(numpy_ver >= [1, 7]):
                    # Needed when we disable the old API, as cython
                    # use the old interface
                    preargs.append("-DNPY_ENSUREARRAY=NPY_ARRAY_ENSUREARRAY")
                    preargs.append("-DNPY_ENSURECOPY=NPY_ARRAY_ENSURECOPY")
                    preargs.append("-DNPY_ALIGNED=NPY_ARRAY_ALIGNED")
                    preargs.append("-DNPY_WRITEABLE=NPY_ARRAY_WRITEABLE")
                    preargs.append("-DNPY_UPDATE_ALL=NPY_ARRAY_UPDATE_ALL")
                    preargs.append("-DNPY_C_CONTIGUOUS=NPY_ARRAY_C_CONTIGUOUS")
                    preargs.append("-DNPY_F_CONTIGUOUS=NPY_ARRAY_F_CONTIGUOUS")

            cmodule.GCC_compiler.compile_str(
                dirname, code, location=loc, preargs=preargs, hide_symbols=False
            )
            # Save version into the __init__.py file.
            init_py = os.path.join(loc, "__init__.py")
            with open(init_py, "w") as f:
                f.write(f"_version = {version}\n")
            # If we just compiled the module for the first time, then it was
            # imported at the same time: we need to make sure we do not
            # reload the now outdated __init__.pyc below.
            init_pyc = os.path.join(loc, "__init__.pyc")
            if os.path.isfile(init_pyc):
                os.remove(init_pyc)
            try_import()

            try_reload()
            from scan_perform import scan_perform as scan_c

            assert scan_perform._version == scan_c.get_version()
            _logger.info(f"New version {scan_perform._version}")
    finally:
        # Release lock on compilation directory.
        release_lock()

# This is caused as cython use the old NumPy C-API but we use the new one.
# To fix it completely, we would need to modify Cython to use the new API.
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="numpy.ndarray size changed")
    from scan_perform.scan_perform import get_version, perform  # noqa: F401
assert version == get_version()



import configparser as ConfigParser
import hashlib
import logging
import os
import shlex
import sys
import typing
import warnings
from functools import wraps
from io import StringIO


_logger = logging.getLogger("theano.configparser")


class TheanoConfigWarning(Warning):
    def warn(cls, message, stacklevel=0):
        warnings.warn(message, cls, stacklevel=stacklevel + 3)

    warn = classmethod(warn)


def parse_config_string(config_string, issue_warnings=True):
    """
    Parses a config string (comma-separated key=value components) into a dict.
    """
    config_dict = {}
    my_splitter = shlex.shlex(config_string, posix=True)
    my_splitter.whitespace = ","
    my_splitter.whitespace_split = True
    for kv_pair in my_splitter:
        kv_pair = kv_pair.strip()
        if not kv_pair:
            continue
        kv_tuple = kv_pair.split("=", 1)
        if len(kv_tuple) == 1:
            if issue_warnings:
                TheanoConfigWarning.warn(
                    f"Config key '{kv_tuple[0]}' has no value, ignoring it",
                    stacklevel=1,
                )
        else:
            k, v = kv_tuple
            # subsequent values for k will override earlier ones
            config_dict[k] = v
    return config_dict


# THEANORC can contain a colon-delimited list of config files, like
# THEANORC=~lisa/.theanorc:~/.theanorc
# In that case, definitions in files on the right (here, ~/.theanorc) have
# precedence over those in files on the left.
def config_files_from_theanorc():
    rval = [
        os.path.expanduser(s)
        for s in os.getenv("THEANORC", "~/.theanorc").split(os.pathsep)
    ]
    if os.getenv("THEANORC") is None and sys.platform == "win32":
        # to don't need to change the filename and make it open easily
        rval.append(os.path.expanduser("~/.theanorc.txt"))
    return rval


class change_flags:
    """
    Use this as a decorator or context manager to change the value of
    Theano config variables.

    Useful during tests.
    """

    def __init__(self, args=(), **kwargs):
        confs = dict()
        args = dict(args)
        args.update(kwargs)
        for k in args:
            l = [v for v in _config_var_list if v.fullname == k]
            assert len(l) == 1, l
            confs[k] = l[0]
        self.confs = confs
        self.new_vals = args

    def __call__(self, f):
        @wraps(f)
        def res(*args, **kwargs):
            with self:
                return f(*args, **kwargs)

        return res

    def __enter__(self):
        self.old_vals = {}
        for k, v in self.confs.items():
            self.old_vals[k] = v.__get__(True, None)
        try:
            for k, v in self.confs.items():
                v.__set__(None, self.new_vals[k])
        except Exception:
            _logger.error(f"Failed to change flags for {self.confs}.")
            self.__exit__()
            raise

    def __exit__(self, *args):
        for k, v in self.confs.items():
            v.__set__(None, self.old_vals[k])


def fetch_val_for_key(key, delete_key=False):
    """Return the overriding config value for a key.
    A successful search returns a string value.
    An unsuccessful search raises a KeyError

    The (decreasing) priority order is:
    - THEANO_FLAGS
    - ~./theanorc

    """

    # first try to find it in the FLAGS
    if key in THEANO_FLAGS_DICT:
        if delete_key:
            return THEANO_FLAGS_DICT.pop(key)
        return THEANO_FLAGS_DICT[key]

    # next try to find it in the config file

    # config file keys can be of form option, or section__option
    key_tokens = key.rsplit("__", 1)
    if len(key_tokens) > 2:
        raise KeyError(key)

    if len(key_tokens) == 2:
        section, option = key_tokens
    else:
        section, option = "global", key
    try:
        try:
            return theano_cfg.get(section, option)
        except ConfigParser.InterpolationError:
            return theano_raw_cfg.get(section, option)
    except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
        raise KeyError(key)


def _config_print(thing, buf, print_doc=True):
    for cv in _config_var_list:
        print(cv, file=buf)
        if print_doc:
            print("    Doc: ", cv.doc, file=buf)
        print("    Value: ", cv.__get__(True, None), file=buf)
        print("", file=buf)


def _hash_from_code(msg):
    """This function was copied from theano.gof.utils to get rid of that import."""
    # hashlib.sha256() requires an object that supports buffer interface,
    # but Python 3 (unicode) strings don't.
    if isinstance(msg, str):
        msg = msg.encode()
    # Python 3 does not like module names that start with
    # a digit.
    return "m" + hashlib.sha256(msg).hexdigest()


def get_config_hash():
    """
    Return a string sha256 of the current config options. In the past,
    it was md5.

    The string should be such that we can safely assume that two different
    config setups will lead to two different strings.

    We only take into account config options for which `in_c_key` is True.
    """
    all_opts = sorted(
        [c for c in _config_var_list if c.in_c_key], key=lambda cv: cv.fullname
    )
    return _hash_from_code(
        "\n".join(
            ["{} = {}".format(cv.fullname, cv.__get__(True, None)) for cv in all_opts]
        )
    )


class TheanoConfigParser:
    # properties are installed by AddConfigVar
    _i_am_a_config_class = True

    def __str__(self, print_doc=True):
        sio = StringIO()
        _config_print(self.__class__, sio, print_doc=print_doc)
        return sio.getvalue()


class ConfigParam:
    """Base class of all kinds of configuration parameters.

    A ConfigParam has not only default values and configurable mutability, but
    also documentation text, as well as filtering and validation routines
    that can be context-dependent.
    """

    def __init__(
        self,
        default: typing.Union[object, typing.Callable[[object], object]],
        *,
        apply: typing.Optional[typing.Callable[[object], object]] = None,
        validate: typing.Optional[typing.Callable[[object], bool]] = None,
        mutable: bool = True,
    ):
        """
        Represents a configuration parameter and its associated casting and validation logic.

        Parameters
        ----------
        default : object or callable
            A default value, or function that returns a default value for this parameter.
        apply : callable, optional
            Callable that applies a modification to an input value during assignment.
            Typical use cases: type casting or expansion of '~' to user home directory.
        validate : callable, optional
            A callable that validates the parameter value during assignment.
            It may raise an (informative!) exception itself, or simply return True/False.
            For example to check the availability of a path, device or to restrict a float into a range.
        mutable : bool
            If mutable is False, the value of this config settings can not be changed at runtime.
        """
        self._default = default
        self._apply = apply
        self._validate = validate
        self._mutable = mutable
        self.is_default = True
        # set by AddConfigVar:
        self.fullname = None
        self.doc = None

        # Note that we do not call `self.filter` on the default value: this
        # will be done automatically in AddConfigVar, potentially with a
        # more appropriate user-provided default value.
        # Calling `filter` here may actually be harmful if the default value is
        # invalid and causes a crash or has unwanted side effects.
        super().__init__()

    @property
    def default(self):
        return self._default

    @property
    def mutable(self) -> bool:
        return self._mutable

    def apply(self, value):
        """Applies modifications to a parameter value during assignment.

        Typical use cases are casting or the subsitution of '~' with the user home directory.
        """
        if callable(self._apply):
            return self._apply(value)
        return value

    def validate(self, value) -> None:
        """Validates that a parameter values falls into a supported set or range.

        Raises
        ------
        ValueError
            when the validation turns out negative
        """
        if not callable(self._validate):
            return True
        if self._validate(value) is False:
            raise ValueError(
                f"Invalid value ({value}) for configuration variable '{self.fullname}'."
            )
        return True

    def __get__(self, cls, type_, delete_key=False):
        if cls is None:
            return self
        if not hasattr(self, "val"):
            try:
                val_str = fetch_val_for_key(self.fullname, delete_key=delete_key)
                self.is_default = False
            except KeyError:
                if callable(self.default):
                    val_str = self.default()
                else:
                    val_str = self.default
            self.__set__(cls, val_str)
        return self.val

    def __set__(self, cls, val):
        if not self.mutable and hasattr(self, "val"):
            raise Exception(
                "Can't change the value of {self.fullname} config parameter after initialization!"
            )
        applied = self.apply(val)
        self.validate(applied)
        self.val = applied


class EnumStr(ConfigParam):
    def __init__(
        self, default: str, options: typing.Sequence[str], validate=None, mutable=True
    ):
        """Creates a str-based parameter that takes a predefined set of options.

        Parameters
        ----------
        default : str
            The default setting.
        options : sequence
            Further str values that the parameter may take.
            May, but does not need to include the default.
        validate : callable
            See `ConfigParam`.
        mutable : callable
            See `ConfigParam`.
        """
        self.all = {default, *options}

        # All options should be strings
        for val in self.all:
            if not isinstance(val, str):
                raise ValueError(f"Non-str value '{val}' for an EnumStr parameter.")
        super().__init__(default, apply=self._apply, validate=validate, mutable=mutable)

    def _apply(self, val):
        if val in self.all:
            return val
        else:
            raise ValueError(
                f"Invalid value ('{val}') for configuration variable '{self.fullname}'. "
                f"Valid options are {self.all}"
            )

    def __str__(self):
        return f"{self.fullname} ({self.all}) "


class TypedParam(ConfigParam):
    def __str__(self):
        # The "_apply" callable is the type itself.
        return f"{self.fullname} ({self._apply}) "


class StrParam(TypedParam):
    def __init__(self, default, validate=None, mutable=True):
        super().__init__(default, apply=str, validate=validate, mutable=mutable)


class IntParam(TypedParam):
    def __init__(self, default, validate=None, mutable=True):
        super().__init__(default, apply=int, validate=validate, mutable=mutable)


class FloatParam(TypedParam):
    def __init__(self, default, validate=None, mutable=True):
        super().__init__(default, apply=float, validate=validate, mutable=mutable)


class BoolParam(TypedParam):
    """A boolean parameter that may be initialized from any of the following:
    False, 0, "false", "False", "0"
    True, 1, "true", "True", "1"
    """

    def __init__(self, default, validate=None, mutable=True):
        super().__init__(default, apply=self._apply, validate=validate, mutable=mutable)

    def _apply(self, value):
        if value in {False, 0, "false", "False", "0"}:
            return False
        elif value in {True, 1, "true", "True", "1"}:
            return True
        raise ValueError(
            f"Invalid value ({value}) for configuration variable '{self.fullname}'."
        )


class DeviceParam(ConfigParam):
    def __init__(self, default, *options, **kwargs):
        super().__init__(
            default, apply=self._apply, mutable=kwargs.get("mutable", True)
        )

    def _apply(self, val):
        if val == self.default or val.startswith("opencl") or val.startswith("cuda"):
            return val
        elif val.startswith("gpu"):
            raise ValueError(
                "You are tring to use the old GPU back-end. "
                "It was removed from Theano. Use device=cuda* now. "
                "See https://github.com/Theano/Theano/wiki/Converting-to-the-new-gpu-back-end%28gpuarray%29 "
                "for more information."
            )
        else:
            raise ValueError(
                'Invalid value ("{val}") for configuration '
                'variable "{self.fullname}". Valid options start with '
                'one of "cpu", "opencl" or "cuda".'
            )

    def __str__(self):
        return f"{self.fullname} ({self.default}, opencl*, cuda*) "


class ContextsParam(ConfigParam):
    def __init__(self):
        super().__init__("", apply=self._apply, mutable=False)

    def _apply(self, val):
        if val == "":
            return val
        for v in val.split(";"):
            s = v.split("->")
            if len(s) != 2:
                raise ValueError(f"Malformed context map: {v}")
            if s[0] == "cpu" or s[0].startswith("cuda") or s[0].startswith("opencl"):
                raise ValueError(f"Cannot use {s[0]} as context name")
        return val


# TODO: Not all of the following variables need to exist. Most should be private.
THEANO_FLAGS = os.getenv("THEANO_FLAGS", "")
# The THEANO_FLAGS environment variable should be a list of comma-separated
# [section__]option=value entries. If the section part is omitted, there should
# be only one section that contains the given option.
THEANO_FLAGS_DICT = parse_config_string(THEANO_FLAGS, issue_warnings=True)
_config_var_list = []

config_files = config_files_from_theanorc()
theano_cfg = ConfigParser.ConfigParser(
    {
        "USER": os.getenv("USER", os.path.split(os.path.expanduser("~"))[-1]),
        "LSCRATCH": os.getenv("LSCRATCH", ""),
        "TMPDIR": os.getenv("TMPDIR", ""),
        "TEMP": os.getenv("TEMP", ""),
        "TMP": os.getenv("TMP", ""),
        "PID": str(os.getpid()),
    }
)
theano_cfg.read(config_files)
# Having a raw version of the config around as well enables us to pass
# through config values that contain format strings.
# The time required to parse the config twice is negligible.
theano_raw_cfg = ConfigParser.RawConfigParser()
theano_raw_cfg.read(config_files)

# N.B. all instances of TheanoConfigParser give access to the same properties.
config = TheanoConfigParser()


def AddConfigVar(name, doc, configparam, root=config, in_c_key=True):
    """Add a new variable to `theano.config`.

    The data structure at work here is a tree of classes with class
    attributes/properties that are either a) instantiated dynamically-generated
    classes, or b) `ConfigParam` instances.  The root of this tree is the
    `TheanoConfigParser` class, and the internal nodes are the ``SubObj``
    classes created inside of `AddConfigVar`.

    Why this design?
    - The config object is a true singleton.  Every instance of
    `TheanoConfigParser` is an empty instance that looks up
    attributes/properties in the [single] ``TheanoConfigParser.__dict__``
    - The subtrees provide the same interface as the root
    - `ConfigParser` subclasses control get/set of config properties to guard
    against craziness.

    This method also performs some of the work of initializing `ConfigParam`
    instances

    Parameters
    ----------
    name: string
        The full name for this configuration variable. Takes the form
        ``"[section0__[section1__[etc]]]_option"``.
    doc: string
        A string that provides documentation for the config variable.
    configparam: ConfigParam
        An object for getting and setting this configuration parameter
    root: object
        Used for recursive calls.  Do not provide a value for this parameter.
    in_c_key: boolean
        If ``True``, then whenever this config option changes, the key
        associated to compiled C modules also changes, i.e. it may trigger a
        compilation of these modules (this compilation will only be partial if it
        turns out that the generated C code is unchanged). Set this option to False
        only if you are confident this option should not affect C code compilation.

    """

    if root is config:
        # Only set the name in the first call, not the recursive ones
        configparam.fullname = name
    if "." in name:
        raise ValueError(
            f"Dot-based sections were removed. Use double underscores! ({name})"
        )
    if hasattr(root, name):
        raise AttributeError(f"The name {configparam.fullname} is already taken")
    configparam.doc = doc
    configparam.in_c_key = in_c_key
    # Trigger a read of the value from config files and env vars
    # This allow to filter wrong value from the user.
    if not callable(configparam.default):
        configparam.__get__(root, type(root), delete_key=True)
    else:
        # We do not want to evaluate now the default value
        # when it is a callable.
        try:
            fetch_val_for_key(configparam.fullname)
            # The user provided a value, filter it now.
            configparam.__get__(root, type(root), delete_key=True)
        except KeyError:
            _logger.error(
                f"Suppressed KeyError in AddConfigVar for parameter '{name}' with fullname '{configparam.fullname}'!"
            )
    setattr(root.__class__, name, configparam)
    # TODO: After assigning the configvar to the "root" object, there should be
    # no reason to keep the _config_var_list around!
    _config_var_list.append(configparam)



"""
Theano is an optimizing compiler in Python, built to evaluate
complicated expressions (especially matrix-valued ones) as quickly as
possible.  Theano compiles expression graphs (see :doc:`graph` ) that
are built by Python code. The expressions in these graphs are called
`Apply` nodes and the variables in these graphs are called `Variable`
nodes.

You compile a graph by calling `function`, which takes a graph, and
returns a callable object.  One of theano's most important features is
that `function` can transform your graph before compiling it.  It can
replace simple expressions with faster or more numerically stable
implementations.

To learn more, check out:

- Op List (:doc:`oplist`)

The markup language used in the docstrings is ReStructured Text,
which may be rendered with Sphinx. A rendered version is
maintained at http://www.deeplearning.net/software/theano/library/

"""


__docformat__ = "restructuredtext en"

# Set a default logger. It is important to do this before importing some other
# theano code, since this code may want to log some messages.
import logging
import os
import sys


theano_logger = logging.getLogger("theano")
logging_default_handler = logging.StreamHandler()
logging_default_formatter = logging.Formatter(
    fmt="%(levelname)s (%(name)s): %(message)s"
)
logging_default_handler.setFormatter(logging_default_formatter)
theano_logger.setLevel(logging.WARNING)

if not theano_logger.hasHandlers():
    theano_logger.addHandler(logging_default_handler)


# Disable default log handler added to theano_logger when the module
# is imported.
def disable_log_handler(logger=theano_logger, handler=logging_default_handler):
    if logger.hasHandlers():
        logger.removeHandler(handler)


# Version information.
from theano.version import version as __version__


# Raise a meaningful warning/error if the theano directory is in the Python
# path.
rpath = os.path.realpath(__path__[0])
for p in sys.path:
    if os.path.realpath(p) != rpath:
        continue
    raise RuntimeError("You have the theano directory in your Python path.")

from theano.configdefaults import config
from theano.configparser import change_flags


# This is the api version for ops that generate C code.  External ops
# might need manual changes if this number goes up.  An undefined
# __api_version__ can be understood to mean api version 0.
#
# This number is not tied to the release version and should change
# very rarely.
__api_version__ = 1

from theano import scalar, tensor
from theano.compile import (
    In,
    Mode,
    Out,
    Param,
    ProfileStats,
    SymbolicInput,
    SymbolicOutput,
    as_op,
    predefined_linkers,
    predefined_modes,
    predefined_optimizers,
    shared,
)
from theano.compile.function import function, function_dump
from theano.compile.function.types import FunctionMaker
from theano.gof import (
    Apply,
    CLinker,
    Constant,
    Container,
    DualLinker,
    FunctionGraph,
    Generic,
    InconsistencyError,
    Linker,
    LocalLinker,
    Op,
    OpenMPOp,
    OpWiseCLinker,
    PerformLinker,
    Type,
    Variable,
    generic,
    object2,
    opt,
    toolbox,
    utils,
)
from theano.gradient import Lop, Rop, grad, subgraph_grad
from theano.misc.safe_asarray import _asarray
from theano.printing import pp, pprint
from theano.updates import OrderedUpdates


if (
    config.device.startswith("cuda")
    or config.device.startswith("opencl")
    or config.init_gpu_device.startswith("cuda")
    or config.init_gpu_device.startswith("opencl")
    or config.contexts != ""
):
    import theano.gpuarray

# Use config.numpy to call numpy.seterr
import numpy as np


if config.numpy__seterr_all == "None":
    _all = None
else:
    _all = config.numpy__seterr_all
if config.numpy__seterr_divide == "None":
    _divide = None
else:
    _divide = config.numpy__seterr_divide
if config.numpy__seterr_over == "None":
    _over = None
else:
    _over = config.numpy__seterr_over
if config.numpy__seterr_under == "None":
    _under = None
else:
    _under = config.numpy__seterr_under
if config.numpy__seterr_invalid == "None":
    _invalid = None
else:
    _invalid = config.numpy__seterr_invalid
np.seterr(all=_all, divide=_divide, over=_over, under=_under, invalid=_invalid)
del _all, _divide, _over, _under, _invalid


def get_scalar_constant_value(v):
    """Return the constant scalar (i.e. 0-D) value underlying variable `v`.

    If v is the output of dimshuffles, fills, allocs, rebroadcasts, cast
    this function digs through them.

    If theano.sparse is also there, we will look over CSM op.

    If `v` is not some view of constant data, then raise a
    tensor.basic.NotScalarConstantError.
    """
    # Is it necessary to test for presence of theano.sparse at runtime?
    sparse = globals().get("sparse")
    if sparse and isinstance(v.type, sparse.SparseType):
        if v.owner is not None and isinstance(v.owner.op, sparse.CSM):
            data = v.owner.inputs[0]
            return tensor.get_scalar_constant_value(data)
    return tensor.get_scalar_constant_value(v)


def sparse_grad(var):
    """This function return a new variable whose gradient will be
    stored in a sparse format instead of dense.

    Currently only variable created by AdvancedSubtensor1 is supported.
    i.e. a_tensor_var[an_int_vector].

    .. versionadded:: 0.6rc4
    """
    assert isinstance(var.owner.op, tensor.AdvancedSubtensor1)
    ret = var.owner.op.__class__(sparse_grad=True)(*var.owner.inputs)
    return ret


import theano.tensor.shared_randomstreams
from theano.scan import checkpoints, clone, foldl, foldr, map, reduce, scan



"""Graph optimization framework"""

import theano
from theano.gof.cc import CLinker, DualLinker, HideC, OpWiseCLinker
from theano.gof.destroyhandler import DestroyHandler
from theano.gof.fg import FunctionGraph, InconsistencyError, MissingInputError
from theano.gof.graph import Apply, Constant, Variable, view_roots
from theano.gof.link import (
    Container,
    Linker,
    LocalLinker,
    PerformLinker,
    WrapLinker,
    WrapLinkerMany,
)
from theano.gof.op import (
    COp,
    Op,
    OpenMPOp,
    PureOp,
    get_test_value,
    ops_with_inner_function,
)
from theano.gof.opt import (
    CheckStackTraceOptimization,
    EquilibriumOptimizer,
    GlobalOptimizer,
    LocalOptGroup,
    LocalOptimizer,
    MergeOptimizer,
    NavigatorOptimizer,
    OpKeyOptimizer,
    OpRemove,
    OpSub,
    PatternSub,
    SeqOptimizer,
    TopoOptimizer,
    inplace_optimizer,
    local_optimizer,
    optimizer,
)
from theano.gof.optdb import DB, EquilibriumDB, LocalGroupDB, ProxyDB, Query, SequenceDB
from theano.gof.params_type import Params, ParamsType
from theano.gof.toolbox import (
    Bookkeeper,
    Feature,
    History,
    NodeFinder,
    NoOutputFromInplace,
    PrintListener,
    ReplacementDidNotRemoveError,
    ReplaceValidate,
    Validator,
)
from theano.gof.type import CEnumType, EnumList, EnumType, Generic, Type, generic
from theano.gof.utils import MethodNotDefined, hashtype, object2


if theano.config.cmodule__preload_cache:
    from theano.gof.cc import get_module_cache

    get_module_cache()
